// (C) Copyright Andrew Bazan, 3/2000
// Feel free to make modifications or use this code in something else,
// just please give me credit where credit is due. Thanks.

#include "BeAlk3.h"

AlkView::AlkView(BRect frame)
	: BView(frame, "AlkView", B_FOLLOW_ALL_SIDES, B_WILL_DRAW|B_FRAME_EVENTS)
{
	running = false;
	legitDown = false;
	frameWidth = 250;
	frameHeight = 220;
	maxIterations = 32;
	paletteType = 1; // mixed 1
	fractalType = 1; // mandelbrot
	myFB = new BBitmap(Bounds(), B_RGB32);
}

AlkView::~AlkView()
{
	delete myFB;
}

void AlkView::NewFractal()
{
	if(fractalType == 1)
	{
		initReal = -1.6;
		initImag = 1.1;
		curHStep = .01 * (250.0 / frameWidth);
		curVStep = .01 * (220.0 / frameHeight);
	}
	else // setup initial julia values
	{
		initReal = -1.9;
		initImag = .9;
		curHStep = .015 * (250.0 / frameWidth);
		curVStep = .015 * (220.0 / frameHeight);
	}
	Fractalize();
	running = true;
	Draw(Bounds());
}

void AlkView::SetIterations(int iterations)
{
	maxIterations = iterations;
}

void AlkView::SetPalette(int8 palette)
{
	paletteType = palette;
}

void AlkView::SetFractal(int8 fractal)
{
	fractalType = fractal;
	if(fractalType == 1)
	{
		curHStep = .01 * (250.0 / frameWidth);
		curVStep = .01 * (220.0 / frameHeight);
	}
	else
	{
		curHStep = .015 * (250.0 / frameWidth);
		curVStep = .015 * (220.0 / frameHeight);
	}
}

void AlkView::Draw(BRect updateRect)
{
	if(running)
		DrawBitmap(myFB, BPoint(0,0));
}

void AlkView::MouseDown(BPoint myPoint)
{
	if(!running)
		return;
	legitDown = true;
	BRect r(myPoint.x, myPoint.y, myPoint.x, myPoint.y);
	BeginRectTracking(r, B_TRACK_RECT_CORNER);
	downX = int(myPoint.x);
	downY = int(myPoint.y);
}

void AlkView::MouseUp(BPoint myPoint)
{
	float leftX, leftY, width, height;
	double widthRatio, heightRatio;

	if(!legitDown)
		return;

	EndRectTracking();
	if(myPoint.x == downX)
		return;
	if(myPoint.y == downY)
		return;

	if(myPoint.x < downX)
	{
		leftX = myPoint.x;
		width = downX - myPoint.x;
	}
	else
	{
		leftX = downX;
		width = myPoint.x - downX;
	}
	if(myPoint.y < downY)
	{
		leftY = myPoint.y;
		height = downY - myPoint.y;
	}
	else
	{
		leftY = downY;
		height = myPoint.y - downY;
	}

	widthRatio = width / frameWidth;
	heightRatio = height / frameHeight;
	
	initReal = initReal + (leftX * curHStep);
	initImag = initImag - (leftY * curVStep);

	curHStep *= widthRatio;
	curVStep *= heightRatio;

	legitDown = false;
	Fractalize();
	Draw(Bounds());
}

void AlkView::FrameResized(float width, float height)
{
	oldWidth = frameWidth;
	oldHeight = frameHeight;
	frameWidth = int(width) + 1;
	frameHeight = int(height) + 1;
	double widthRatio = double(oldWidth) / frameWidth;
	double heightRatio = double(oldHeight) / frameHeight;
	
	if(myFB)
		delete myFB;
	
	myFB = new BBitmap(Bounds(), B_RGB32);

	curHStep *= widthRatio;
	curVStep *= heightRatio;
	
	if(running)
	{
		Fractalize();
		Draw(Bounds());
	}
}

void AlkView::Fractalize()
{
	double curReal = initReal, curImag = initImag;
	uchar *bits = (uchar *)myFB->Bits();
	curReal = initReal;

	for(int y = 0; y < frameHeight; ++y)
	{
		for(int x = 0; x < frameWidth; ++x)
		{
			int16 r, g, b, iterations;
			curReal += curHStep;
			FastComplex c(curReal, curImag);
			if(fractalType == 1)
				iterations = Iterate(c);
			else
				iterations = IterateJulia(c);
			GetColor(iterations, r, g, b);
			*bits++ = g;
			*bits++ = b;
			*bits++ = r;
			bits++; // alpha, skip
		}
		if(y % 20 == 0)
			DrawBitmap(myFB, BPoint(0,0));
		curImag -= curVStep;
		curReal = initReal;
	}
}

int16 AlkView::Iterate(const FastComplex &c)
{
	FastComplex z;
	int iterations = 0;
	
	while(z.AbsoluteValue() < 2 && iterations < maxIterations)
	{
		z = z.Add(c);
		z = z.Square();
		++iterations;
	}

	return iterations;
}

int16 AlkView::IterateJulia(FastComplex z)
{
	FastComplex c(0, .77); // arbitrary numbers, they look good
	int iterations = 0;
	
	while(z.AbsoluteValue() < 2 && iterations < maxIterations)
	{
		z = z.Add(c);
		z = z.Square();
		++iterations;
	}
	
	return iterations;
}

// this color scheme is questionable, at best. first off, calculating this
// on the fly tends to slow things down, and second, as you raise the
// number of max iterations things tend to get darker & darker.
void AlkView::GetColor(const int16 &iterations, int16 &r, int16 &g, int16 &b)
{
	float itRatio = float(iterations) / maxIterations;
	int theCheck = int(itRatio * 100);
	
	if(iterations == maxIterations)
	{
		r = 0;
		g = 0;
		b = 0;

		return;
	}
	
	if(paletteType == 1)
	{
		if(theCheck > 50)
		{
			r = short(255 * itRatio);
			g = short(itRatio * 75);
			b = short(itRatio * 75);
		}
		else if(theCheck > 25)
		{
			r = short(itRatio * 75);
			g = short(500 * itRatio);
			b = short(itRatio * 75);
		}
		else
		{
			r = short(itRatio * 75);
			g = short(itRatio * 75);
			b = short(1000 * itRatio);
		}
	}
	else if(paletteType == 2)
	{
		if(theCheck > 50)
		{
			r = short(itRatio * 75);
			g = short(itRatio * 75);
			b = short(255 * itRatio);
		}
		else if(theCheck > 25)
		{
			r = short(itRatio * 75);
			g = short(500 * itRatio);
			b = short(itRatio * 75);
		}
		else
		{
			r = short(255 * itRatio);
			g = short(itRatio * 75);
			b = short(itRatio * 75);
		}
	}
	else if(paletteType == 3)
	{
		if(theCheck > 50)
		{
			r = short(itRatio * 75);
			g = short(255 * itRatio);
			b = short(itRatio * 75);
		}
		else if(theCheck > 25)
		{
			r = short(255 * itRatio);
			g = short(itRatio * 75);
			b = short(itRatio * 75);
		}
		else
		{
			r = short(itRatio * 75);
			g = short(itRatio * 75);
			b = short(255 * itRatio);
		}
	}
	else if(paletteType == 4)
	{
		r = theCheck * 2;
		b = theCheck / 3;
		g = theCheck / 3;
	}
	else if(paletteType == 5)
	{
		r = theCheck / 3;
		b = theCheck * 2;
		g = theCheck / 3;
	}
	else if(paletteType == 6)
	{
		r = theCheck / 3;
		b = theCheck / 3;
		g = theCheck * 2;
	}
}